/**********************************************************************************

	AUTHOR: Erlend Halsteinli
	CREATED: SPRING 2009 @ IET NTNU

	Class performing decoding of JPEG2000 codestream data on CPU with Kakadu.

***********************************************************************************/

#include "HostDecoder.h"

//Contructor
//Parameters:
//codestreamBuffer [in] - pointer to pointer to buffer with compressed codestream.
//bytesInBuffer [in] - pointer to pointer to the number of bytes in the compressed codestreams.
//Return:
HostDecoder::HostDecoder(char** codestreamBuffer[], int** bytesInBuffer){

	remove("decoder_debug_log.txt");
	debugLog.open("decoder_debug_log.txt",ios_base::out);

	//Copying pointers to codestreamBuffer and bytesInBuffer.
	for(int i=0;i<CODESTREAM_BUFFER_LENGTH;i++){
		jpcBuffer[i]=*(codestreamBuffer[i]);
	}
	bytesInJpcBuffer=*bytesInBuffer;
	
	//initiating codestream buffer object
	codestreamSource.init("crowd_720p_50Mb.jpc");

	// Custom Kakadu messaging services
	kdu_customize_warnings(&pretty_cout);
	kdu_customize_errors(&pretty_cerr);

	//initiating codestream object
	codestream.create(&codestreamSource);
	codestream.enable_restart();

	stripeHeights[0]=imgHeight;
	stripeHeights[1]=imgHeight;
	stripeHeights[2]=imgHeight;

	sampleGaps[0]=1;
	sampleGaps[1]=1;
	sampleGaps[2]=1;

	envDbufHeight=NULL;

	//Add more threads to kakadu if processor has multiple cores. e.g. add n-1 threads to n a core processor.
	env.create();
	env.add_thread();
	env_ref = &env;

	//env_ref=NULL;

	//Allocated one array for each component, but all arrays are subparts of one big array: faster host -> device transfer.
	hostOutput0[0] = (unsigned char *) AllocateMemHost(imgWidth*imgHeight*2);
	hostOutput0[1] = (hostOutput0[0]+imgWidth*imgHeight);
	hostOutput0[2] = (hostOutput0[0]+imgWidth*imgHeight*3/2);

	hostOutput1[0] = (unsigned char *) AllocateMemHost(imgWidth*imgHeight*2);
	hostOutput1[1] = (hostOutput1[0]+imgWidth*imgHeight);
	hostOutput1[2] = (hostOutput1[0]+imgWidth*imgHeight*3/2);
	

	//Create mutex for kduOutput
	hostOutput0Mutex = CreateMutex(NULL, false,(LPCTSTR)"hostOutput0Mutex");
	if(hostOutput0Mutex==NULL){
		DWORD error = GetLastError();
		debugLog<<"Failed creating hostOutput0Mutex mutex with error code 0x"<<hex<<error<<"."<<endl;
	}

	hostOutput1Mutex = CreateMutex(NULL, false,(LPCTSTR)"hostOutput1Mutex");
	if(hostOutput1Mutex==NULL){
		DWORD error = GetLastError();
		debugLog<<"Failed creating hostOutput1Mutex mutex with error code 0x"<<hex<<error<<"."<<endl;
	}

	//Opening semaphore and mutexes to codestreamBuffers
	codestreamBufferSemaphore = OpenSemaphore(STANDARD_RIGHTS_ALL, false,(LPCTSTR)"codestreamBufferSemaphore");
	if(codestreamBufferSemaphore==NULL){
		DWORD error = GetLastError();
		debugLog<<"Failed opening codestreamBufferSemaphore with error code: 0x"<<hex<<error<<"."<<endl;
		exit(-1);
	}
	codestreamBuff0 = OpenMutex(STANDARD_RIGHTS_ALL, false, (LPCTSTR)"codestreamBuff0");
	if(codestreamBuff0==NULL){
		DWORD error = GetLastError();
		debugLog<<"Failed opening codestreamBuff0 with error code: 0x"<<hex<<error<<"."<<endl;
		exit(-1);
	}
	codestreamBuff1 = OpenMutex(STANDARD_RIGHTS_ALL, false, (LPCTSTR)"codestreamBuff1");
	if(codestreamBuff1==NULL){
		DWORD error = GetLastError();
		debugLog<<"Failed opening codestreamBuff1 with error code: 0x"<<hex<<error<<"."<<endl;
		exit(-1);
	}

	//Initiating GUI, passing pointers to kduOutput so gui know where on host to copy (from host to device) the kduOutput from.
	//InitGui(hostOutput0, hostOutput1, deviceInput, deviceIntermediate, imgWidth, imgHeight);
	InitGui(hostOutput0, hostOutput1, imgWidth, imgHeight);
}//end Jpeg2000Decoder()

//Method for CPU decoding of codestream, using Kakadu stripedecoder.
//Parameters:
//bufferNumber [in] - desides which output buffer to decode to, must be zero or one.
//Return: true if decoding is succesfull, false otherwise.
bool HostDecoder::kduDecodeFrame(int bufferNumber){

	//restarting codestream with memory location containing jpc code stream
	codestream.restart(&codestreamSource);
	//codestream.restart(&codestreamNetwork);

	//The code-stream management machinery makes a serious attempt to identify compliance violations and generates an appropriate terminal error message if it finds one
	//codestream.set_fussy();
	//The code-stream management machinery makes a serious attempt to both detect and recover from errors in the code-stream
	//codestream.set_resilient();
	//Compliance is assumed and checking is minimized; there is no attempt to recover from any errors (default)
	//codestream.set_fast();

	//Choosing the number of components to decode++
	codestream.apply_input_restrictions(0,3,0,0,NULL);

	//Initialize decompressor object
	//Last argument: 8, 16, 32
	decompressor.start(codestream, false, true, env_ref, NULL, envDbufHeight);
	//pull components into individual arrays
	if(bufferNumber == 0){
		decompressor.pull_stripe((kdu_byte**)hostOutput0,stripeHeights, sampleGaps, NULL, NULL);
	}
	else if(bufferNumber == 1){
		decompressor.pull_stripe((kdu_byte**)hostOutput1,stripeHeights, sampleGaps, NULL, NULL);
	}
	else{
		debugLog<<"Invalid buffer number passed to kduDecodeField(): "<<bufferNumber<<endl;
		debugLog<<"No decoding will be performed"<<endl;
		return false;
	}
	//Must call finish() because we reuse the decompressor object.
	decompressor.finish();

	return true;
}//end kduDecodeFrame()

//Method for performance testing the decoder.
//Parameters:
//
//Return: void
void HostDecoder::RunPerformanceTest(){

	fprintf(stderr, "Testing decoder performance...\n");
	int  loopCount;
	clock_t startClk, endClk;
	float loopTime;

	//Only Kakadu decoding
	loopCount = 1200;
	startClk = clock();
	for(int i=0; i<loopCount;i++){
		kduDecodeFrame(0);
	}
	endClk=clock();
	loopTime = ((float)(endClk-startClk))/CLOCKS_PER_SEC;
	fprintf(stderr, "Kakadu decoding loop done, decoded %i frames in %f seconds. DbufHeight: %i\n",loopCount, loopTime, envDbufHeight);
	fprintf(stderr, "fps: %f \n",((float)loopCount)/loopTime);
	SaveKduOutput();

	//Only GPU decoding
	loopCount = 8000;
	startClk = clock();
	for(int i=0; i<loopCount;i++){
		DecodeFrameOnDevice(0);
	}
	endClk=clock();
	loopTime = ((float)(endClk-startClk))/CLOCKS_PER_SEC;
	fprintf(stderr, "GPU decoding loop done, decoded %i frames in %f seconds\n",loopCount, loopTime);
	fprintf(stderr, "fps: %f \n",((float)loopCount)/loopTime);
	fprintf(stderr, "Done testing decoder performance.\n\n");
	
}//end RunPerformanceTest()

//Method used to set parameter used in CPU decoder.
//Parameters:
//newDbufHeight [in] - the new Dbuffer height
//Return: void
void HostDecoder::SetDbufHeight(int newDbufHeight){

	envDbufHeight=newDbufHeight;

}//end SetDbufHeight()

//Method used during debugging.
//Parameters:
//
//Return: void.
void HostDecoder::RunCorrectnessTest(){
	kduDecodeFrame(0);
	SaveKduOutput();

	//DecodeFrameOnDevice(0);
	//writeImage("gpu.pgm",(kdu_byte *) hostOutput0[0], 1, imgHeight, imgWidth, true, true);
	//writeImage("gpu.raw",(kdu_byte *) hostOutput0[0], 1, imgHeight, imgWidth, true, false);

}//end RunCorrectnessTest()

//Start method when forking new thread.
//Parameters:
//
//Return: void.
void HostDecoder::Execute(){

	clock_t startClk = 0, endClk;
	float loopTime;
	int decodeCounter = 0, counterLimit = 500;

	DWORD waitReturn;
	waitReturn = WaitForSingleObject(hostOutput0Mutex, 1000);
	if(waitReturn!=WAIT_OBJECT_0) {
		debugLog<<"Wait for hostOutput0Mutex failed in main loop with error code 0x"<<hex<<waitReturn<<"."<<endl;
		if(waitReturn == WAIT_FAILED){
			DWORD lastError = GetLastError();
			debugLog<<"GetLastError() returned: 0x"<<hex<<lastError<<endl;
		}
		debugLog<<"Exiting..."<<endl;
		exit(-1);
	}
	//This semaphore is signaled once by the packetbuffer2codestream class, when access first is given to codestreamBUff0
	//this ensures that the packetbuffer2codestream class has started to fill up the codestream buffer before this thread try to decode the codestream.
	debugLog<<"Waiting for codestreamBufferSemaphore"<<endl;
	waitReturn = WaitForSingleObject(codestreamBufferSemaphore, 10000);
	if(waitReturn!=WAIT_OBJECT_0) {
		debugLog<<"Wait for codestreamBufferSemaphore failed before thread main loop with error code 0x"<<hex<<waitReturn<<"."<<endl;
		if(waitReturn == WAIT_FAILED){
			DWORD lastError = GetLastError();
			debugLog<<"GetLastError() returned: 0x"<<hex<<lastError<<endl;
		}
		debugLog<<"Exiting.."<<endl;
		exit(-1);
	}

	debugLog<<"Entering thread main loop."<<endl;

	//main CPU decoder loop.
	while(true){

		////Decode into buffer zero:
		//Need access to codestreambuffer
		waitReturn = WaitForSingleObject(codestreamBuff0, 1000);
		if(waitReturn==WAIT_OBJECT_0){
			codestreamSource.update(jpcBuffer[0], *(bytesInJpcBuffer));
			kduDecodeFrame(0);
			//done with codestream buffer, release it so PacketBuffer2Codestream class can fill it again
			ReleaseMutex(codestreamBuff0);
			//new field done with CPU decoding, increase frame rate counter 
			decodeCounter++;

			waitReturn = WaitForSingleObject(hostOutput1Mutex, 1000);
			if(waitReturn==WAIT_OBJECT_0){
				//Obtained access to kduOutputOnHost1 buffer, release kduOutputOnHost0.
				ReleaseMutex(hostOutput0Mutex);
			}
			else{
				debugLog<<"Timeout or error in WaitForSingleObject() on hostOutput1Mutex: "<<waitReturn<<endl;
				debugLog<<"Exiting..."<<endl;
				exit(-1);
			}
		}
		else{
			debugLog<<"Timeout or error in WaitForSingleObject() on codestreamBuff0 mutex: "<<waitReturn<<endl;
			debugLog<<"Exiting..."<<endl;
			exit(-1);
		}////Done with buffer zero

		////Decode into buffer one:
		//Need access to codestreambuffer
		waitReturn = WaitForSingleObject(codestreamBuff1, 1000);
		if(waitReturn==WAIT_OBJECT_0){
			codestreamSource.update(jpcBuffer[1], *(bytesInJpcBuffer+1));
			kduDecodeFrame(1);
			//done with codestream buffer, release it so PacketBuffer2Codestream class can fill it again
			ReleaseMutex(codestreamBuff1);
			//new field done with CPU decoding, increase frame rate counter 
			decodeCounter++;

			waitReturn = WaitForSingleObject(hostOutput0Mutex, 1000);
			if(waitReturn==WAIT_OBJECT_0){
				//Obtained access to kduOutputOnHost0 buffer, release kduOutputOnHost1.
				ReleaseMutex(hostOutput1Mutex);
			}
			else{
				debugLog<<"Timeout or error in WaitForSingleObject() on kduOutputOnHost0 mutex: "<<waitReturn<<endl;
				debugLog<<"Exiting..."<<endl;
				exit(-1);
			}

		}
		else{
			debugLog<<"Timeout or error in WaitForSingleObject() on codestreamBuff1 mutex: "<<waitReturn<<endl;
			debugLog<<"Exiting..."<<endl;
			exit(-1);
		}////Done with buffer one
		
		//tracing CPU decoding rate
		if(decodeCounter>=counterLimit){
			endClk = clock();
			loopTime = ((float)(endClk-startClk))/CLOCKS_PER_SEC;
			fprintf(stderr, "Kakadu decoded %i frames in %f seconds. DbuffHeigth: %i\n", decodeCounter, loopTime, envDbufHeight);
			fprintf(stderr, "fps: %f \n",((float)decodeCounter)/loopTime);
			decodeCounter=0;
			startClk = clock();
		}
	}
}//end Execute()

//Deinterleaving wavelet coeficents to "visual ordering", i.e. LL upper left corner, HL upper rigth corner and so on.
//Not used since it can be done faster on GPU.
void HostDecoder::DeinterleaveVisual(unsigned char * hostLocal[3], int scale, int dwtLevel){

	int stepSize, dstIdx;
	int length;

	//Deinterleaving HL1, LH1 and HH1
	if(dwtLevel>=1){
		stepSize=2;
		//HL1
		dstIdx=imgWidth/stepSize;
		length=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//LH1
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//HH1
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize+imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
	}
	//HL2, LH2 and HH2
	if(dwtLevel>=2){
		stepSize=4;
		//HL2
		length=0;
		dstIdx=imgWidth/stepSize;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//LH2
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//HH2
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize+imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
	}
	//HL3, LH3 and HH3
	if(dwtLevel>=3){
		stepSize=8;
		//HL3
		length=0;
		dstIdx=imgWidth/stepSize;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//LH3
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//HH3
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize+imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
	}
	//HL4, LH4 and HH4
	if(dwtLevel>=4){
		stepSize=16;
		//HL4
		length=0;
		dstIdx=imgWidth/stepSize;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//LH4
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//HH4
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize+imgWidth/stepSize;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
	}
	//HL5, LH5 and HH5. The LH5 and HH5 subband will contain one less sample line(i.e. 40 less samples, which is subband width), than the LL5 and HL5 subband.
	if(dwtLevel>=5){
		stepSize=32;
		//HL5
		length=0;
		dstIdx=imgWidth/stepSize;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//LH5
		length=0;
		dstIdx = (imgHeight*imgWidth)/stepSize+imgWidth/2;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
		//HH5
		length=0;
		dstIdx = imgHeight*imgWidth/stepSize+imgWidth/stepSize+imgWidth/2;
		for(int curRow=stepSize/2; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=stepSize/2; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol)-scale;
				length++;
			}
		}
	}
	//LL1
	if(dwtLevel==1){
		stepSize=2;
		dstIdx=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth/2)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				//LL1
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol);
			}
		}
	}
	//LL2
	else if(dwtLevel==2){
		stepSize=4;
		dstIdx=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				//LL2
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol);
			}
		}
	}
	//LL3
	else if(dwtLevel==3){
		stepSize=8;
		dstIdx=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				//LL3
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol);
			}
		}
	}
	//LL4
	else if(dwtLevel==4){
		stepSize=16;
		dstIdx=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				//LL4
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol);
			}
		}
	}
	//LL5
	else if(dwtLevel==5){
		stepSize=32;
		dstIdx=0;
		for(int curRow=0; curRow<imgHeight; curRow+=stepSize, dstIdx+=(imgWidth-imgWidth/stepSize)){
			for(int curCol=0; curCol<imgWidth; curCol+=stepSize, dstIdx++){
				//LL5
				*(hostLocal[0]+dstIdx)=*(hostOutput0[0]+curRow*imgWidth+curCol);
			}
		}
	}
}//end Deinterleave()

//Method used to save the output from Kakadu to file.
//Parameters:
//
//Return: void
void HostDecoder::SaveKduOutput(){

	bool kakaduIDWTbypassed = true;
	
	if(kakaduIDWTbypassed){
		writeImage("preDeinterleaving.pgm",(kdu_byte *) hostOutput0[0], 1, imgHeight, imgWidth, true, true);

		unsigned char * hostLocal[3];
		hostLocal[0] = (unsigned char*) AllocateMemHost(imgWidth*imgHeight*2);
		hostLocal[1] = (hostLocal[0]+imgWidth*imgHeight);
		hostLocal[2] = (hostLocal[0]+imgWidth*imgHeight*3/2);

		int level = 1;
		DeinterleaveVisual(hostLocal, 0x00, level);
		writeImage("deinterleavedVisual.pgm",(kdu_byte *) hostLocal[0], 1, imgHeight, imgWidth, true, true);


		DeinterleaveVisual(hostLocal, 0x80, level);
		writeImage("deinterleavedVisualScaled.pgm",(kdu_byte *) hostLocal[0], 1, imgHeight, imgWidth, true, true);

		//Deinterleaving wavelet coeficents to "linear ordering", i.e. first all LL coefs, then HL, LH and last HH.
		//processing even indexed rows
		//Not used since it can be done fasten on GPU.
		int destIdx = 0;
		for(int row=0; row<720; row+=2){
			for(int col=0; col<640; col++){
				//LL
				*(hostLocal[0]+destIdx)=*(hostOutput0[0]+row*1280+2*col);
				//HL
				*(hostLocal[0]+640*360+destIdx)=*(hostOutput0[0]+row*1280+2*col+1)-0x0;
				destIdx++;
			}
		}
		destIdx=0;
		//processing odd indexed rows
		for(int row=1; row<720; row+=2){
			for(int col=0; col<640; col++){
				//LH
				*(hostLocal[0]+1280*360+destIdx)=*(hostOutput0[0]+row*1280+2*col)-0x0;
				//HH
				*(hostLocal[0]+1280*360+640*360+destIdx)=*(hostOutput0[0]+row*1280+2*col+1)-0x0;
				destIdx++;
			}
		}

		writeImage("deinterleavedLinear.raw",(kdu_byte *) hostLocal[0], 1, imgHeight, imgWidth, true, false);
		writeImage("deinterleavedLinear.pgm",(kdu_byte *) hostLocal[0], 1, imgHeight, imgWidth, true, true);
	}
	else{
		writeImage("hostOutputY.pgm",(kdu_byte *) hostOutput0[0], 1, imgHeight, imgWidth, true, true);
		writeImage("hostOutputCb.pgm",(kdu_byte *) hostOutput0[1], 1, imgHeight, imgWidth/2, true, true);
		writeImage("hostOutpuCrY.pgm",(kdu_byte *) hostOutput0[1], 1, imgHeight, imgWidth/2, true, true);
	}


}//end SaveKduOutput()