/**********************************************************************************

	AUTHOR: Erlend Halsteinli
	CREATED: SPRING 2009 @ IET NTNU

	Library performing decoding on GPU with CUDA and GUI with OpenGL.

***********************************************************************************/

//#include <stdlib.h>
//#include <stdio.h>
//#include <string.h>
//#include <math.h>

#include <time.h>
#include <windows.h>

#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_runtime_api.h>
#include <cutil_inline.h>
#include <cuda_gl_interop.h>

#include "CudaLib.h"
#include "Ycc2Rgba.cu"
#include "Deinterleave.cu"

extern "C" void Display();
extern "C" void DrawFrame();
extern "C" void Reshape(int x, int y);
extern "C" void Idle();
extern "C" void CleanupGL();

GLuint pbo;
GLuint textureId;
GLuint shader;
int counterLimit = 100;
int framesDisplayed = 0;
time_t startClk, endClk;
float fps=0, fpsDecoded=0;
int imgWidth, imgHeight;
HANDLE hostOutput0Mutex, hostOutput1Mutex;
DWORD waitReturn;
unsigned int *mappedPbo;
int currentHostBuffer;
unsigned char *hostOutput0[3], *hostOutput1[3];
unsigned char *deviceInput[3], *deviceIntermediate0[3];
short* deviceIntermediate1[3];

// shader for displaying floating-point texture
static const char *shader_code = 
"!!ARBfp1.0\n"
"TEX result.color, fragment.texcoord, texture[0], 2D; \n"
"END";

//Decoder method for GPU, with the following steps:
//1. Transfer Kakadu output from host to device
//2. Deinterleaving DWT coefficients
//3. Performing inline inverse DWT
//4. Performing YCRCB->RGBA, and in the same process writing the RGBA data to a Pixel Buffer Object (PBO) mapped to the CUDA memory space. The PBO reside in graphics memory. The PBO is later used by the OpenGL GUI to draw content to display.
//Parameters:
//bufferNumber [in] - which host buffer to continue decoding from.
//Return: void
extern "C" void DecodeFrameOnDevice(int hostBufferNumber){

	//1.step
	if(hostBufferNumber==0){
		//Host output is 16 bit per sample per component
		//cudaMemcpyAsync(deviceInput[0], hostOutput0[0], imgWidth*imgHeight*4, cudaMemcpyHostToDevice, NULL);
		//Host output is 8 bit per sample per component
		cudaMemcpyAsync(deviceInput[0], hostOutput0[0], imgWidth*imgHeight*2, cudaMemcpyHostToDevice, NULL);
		ReleaseMutex(hostOutput0Mutex);
		currentHostBuffer=1;
	}
	else if(hostBufferNumber==1){
		//Host output is 16 bit per sample per component
		//cudaMemcpyAsync(deviceInput[0], hostOutput1[0], imgWidth*imgHeight*4, cudaMemcpyHostToDevice, NULL);
		//Host output is 8 bit per sample per component
		cudaMemcpyAsync(deviceInput[0], hostOutput1[0], imgWidth*imgHeight*2, cudaMemcpyHostToDevice, NULL);
		ReleaseMutex(hostOutput1Mutex);
		currentHostBuffer=0;
	}
	else{
		fprintf(stderr, "Invalid bufferNumber passed to DecodeFrameOnDevice(): %i", hostBufferNumber);
	}

	//2.step
	//int dwtLevel = 5;
	////Y
	//Deinterleave((ushort4*)deviceInput[0], deviceIntermediate0[0], imgWidth, imgHeight, 0x0, dwtLevel);
	////Cb
	//Deinterleave((ushort4*)deviceInput[1], deviceIntermediate0[1], imgWidth/2, imgHeight, 0x0, dwtLevel);
	////Cr
	//Deinterleave((ushort4*)deviceInput[2], deviceIntermediate0[2], imgWidth/2, imgHeight, 0x0, dwtLevel);

	//3.step
	//Replace with IDWT

	//Copying back to host for verification, only in debug
	//cudaMemcpy(hostOutput0[0], deviceIntermediate0[0], imgWidth*imgHeight*2, cudaMemcpyDeviceToHost);

	//4.step
	//Mapping PBO to CUDA memory space
	cudaGLMapBufferObject((void**)&mappedPbo, pbo);
	//Performing ICT
	//Ycc2Rgba(mappedPbo, deviceIntermediate0[0], deviceIntermediate0[1], deviceIntermediate0[2], imgWidth, imgHeight);
	Ycc2Rgba(mappedPbo, deviceInput[0], deviceInput[1], deviceInput[2], imgWidth, imgHeight);
	//Unmap PBO from CUDA memory space
	cudaGLUnmapBufferObject(pbo);

}//end DecodeFrameOnDevice()

//Allocate memory on device
//Parameters:
//numChars [in] - Number of bytes to allocate
//Return: pointer to the allocated memory
extern "C" void* AllocateMemDevice(unsigned int numBytes){
 
	void *bufferOnDevice;
	//cutilSafeCall(cudaMalloc((void**) &bufferOnDevice, numChars));
	cudaMalloc((void**) &bufferOnDevice, numBytes);

return bufferOnDevice;
}//end AllocateMemDevice

//Allocate memory on host
//Parameters:
//numChars [in] - Number of bytes to allocate
//Return: pointer to the allocated memory
extern "C" void* AllocateMemHost(unsigned int numBytes){
 
	void *bufferOnHost;
	cudaMallocHost((void**) &bufferOnHost, numBytes);

return bufferOnHost;
}//end AllocateMemHost()


//Free memory allocated with CUDA
//Parameters:
//data [in] - pointer to memory structure
//Return: void
extern "C" void FreeMem(void* data){

	cudaFree(data);
}//end FreeMem()

//Clean up CUDA
//Parameters:
//
//Return: void
extern "C" void Cleanup(void){

	cudaThreadExit();
}//end Cleanup()

//Set memory on device
//Parameters:
//buffer [in] - pointer to memory structure
//numBytes [in] - number of bytes to set
//Return: void
extern "C" void MemsetDevice(void* buffer, unsigned int numBytes){

	cudaMemset(buffer, 0x80, numBytes);

}//end MemsetDevice()

//Method used to determine decoder speed 
//Parameters
//
//Return: void
extern "C" void PerformanceTest(){

	if(currentHostBuffer==0){
		//Waiting for single access to host buffer one 
		waitReturn = WaitForSingleObject(hostOutput0Mutex, 1000);
	}
	else if(currentHostBuffer==1){
		//Waiting for single access to host buffer one 
		waitReturn = WaitForSingleObject(hostOutput1Mutex, 1000);
	}
	if(waitReturn==WAIT_OBJECT_0){
		//transfering from host to device, performing inverse DWT and YCC->RGBA and write to PBO.
		DecodeFrameOnDevice(currentHostBuffer);
	}
	else{
		fprintf(stderr, "Error or timeout in performanceTest(): %i, currentHostBuffer: %i", waitReturn, currentHostBuffer);
		exit(-1);
	}
}//end PerformanceTest()

//Method called from the GLUT mainloop, performing the GPU decoding and displays the results using OpenGL.
//CUDA kernel which perform YCC->RGB and writes the result to a PBO mapped into CUDA memory. This reduces HOST<->DEVICE memory transfers.
//Parameters:
//
//Return: void
extern "C" void Display(){

    glClear(GL_COLOR_BUFFER_BIT);

	if(currentHostBuffer==0){
		//Waiting for single access to host buffer one 
		waitReturn = WaitForSingleObject(hostOutput0Mutex, 100);
	}
	else if(currentHostBuffer==1){
		//Waiting for single access to host buffer one 
		waitReturn = WaitForSingleObject(hostOutput1Mutex, 100);
	}
	if(waitReturn==WAIT_OBJECT_0){
		//transfering from host to device, performing inverse DWT and YCC->RGBA and write to PBO.
		DecodeFrameOnDevice(currentHostBuffer);
		
		// load texture from pbo
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
		glBindTexture(GL_TEXTURE_2D, textureId);
		//glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imgWidth, imgHeight, GL_ABGR_EXT, GL_UNSIGNED_BYTE, 0);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imgWidth, imgHeight, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
		
		//draw texture to screen
		DrawFrame();
		framesDisplayed++;
	}
	else if(waitReturn==WAIT_TIMEOUT){
		//bind to texture
		glBindTexture(GL_TEXTURE_2D, textureId);
		//draw texture to screen
		DrawFrame();
		framesDisplayed++;
	}
	else{
		fprintf(stderr, "Wait for kduOutputOnHost1 mutex failed in display() with error code: 0x%x \n", waitReturn);
		if(waitReturn == WAIT_FAILED){
			DWORD lastError = GetLastError();
			fprintf(stderr, "GetLastError() returned: 0x%x \n", lastError);
		}
		exit(-1);
	}

	if(framesDisplayed>=counterLimit){
		endClk=clock();
		fps = (float)(framesDisplayed*CLOCKS_PER_SEC)/(float)(endClk-startClk);
		char title[256];
		sprintf(title, "JPEG2000 Video Decoder  %3.1f fps", fps); 
		glutSetWindowTitle(title);
		startClk=endClk;
		framesDisplayed=0;
	}
}//end Display()

//Helper method called from Display()
//Parameters:
//
//Return: void
extern "C" void DrawFrame(){

	// fragment program is required to display floating point texture
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, shader);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	glDisable(GL_DEPTH_TEST);

	glBegin(GL_QUADS);
	glVertex2f(0, 0); glTexCoord2f(0, 0);
	glVertex2f(0, 1); glTexCoord2f(1, 0);
	glVertex2f(1, 1); glTexCoord2f(1, 1);
	glVertex2f(1, 0); glTexCoord2f(0, 1);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);

	glutSwapBuffers();

	glutPostRedisplay();
}//end DrawFrame()

//Idle method called from glutMainLoop();
//Parameters:
//
//Return: void
extern "C" void Idle(){

    glutPostRedisplay();
}//end Idle()

//Method used by OpenGL to resize the vindow content.
//Parameters:
//
//Return: void
extern "C" void Reshape(int x, int y)
{
    glViewport(0, 0, x, y);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0); 
}//end Reshape()

//Clean up OpenGL
//Parameters:
//
//Return: void
extern "C"  void CleanupGL(){

	cutilSafeCall(cudaGLUnregisterBufferObject(pbo));    
	glDeleteBuffersARB(1, &pbo);
    glDeleteTextures(1, &textureId);
    glDeleteProgramsARB(1, &shader);
}//end CLeanupGL()

//Shader used by GUI.
//Parameters:
//
//Return:
extern "C" GLuint compileASMShader(GLenum program_type, const char *code){

    GLuint program_id;
    glGenProgramsARB(1, &program_id);
    glBindProgramARB(program_type, program_id);
    glProgramStringARB(program_type, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei) strlen(code), (GLubyte *) code);

    GLint error_pos;
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &error_pos);
    if (error_pos != -1) {
        const GLubyte *error_string;
        error_string = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
        fprintf(stderr, "Program error at position: %d\n%s\n", (int)error_pos, error_string);
        return 0;
    }
    return program_id;
}//end compileASMShader()

//Initiate OpenGL
//Parameters:
//
//Return: void
extern "C" void InitOpenGL(){

	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    // create pixel buffer object
    glGenBuffersARB(1, &pbo);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
	glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, imgWidth*imgHeight*sizeof(GLubyte)*4, NULL, GL_STREAM_DRAW_ARB);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
	//cudaGLRegisterBufferObject(pbo);
	cutilSafeCall(cudaGLRegisterBufferObject(pbo));

    // create texture for display
    glGenTextures(1, &textureId);
    glBindTexture(GL_TEXTURE_2D, textureId);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, imgWidth, imgHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glBindTexture(GL_TEXTURE_2D, 0);

    //load shader program
    shader = compileASMShader(GL_FRAGMENT_PROGRAM_ARB, shader_code);
}//end InitOpenGL()

//Initiate GUI and GPU decoder.
//Parameters:
//kduOutputH0 [in] - pointer to host decoder output buffer0
//kduOutputH1 [in] - pointer to host decoder output buffer1
//width [in] - horizontal frame dimension
//height [in] - vertical frame dimension
//Return: void
extern "C" void InitGui(unsigned char *hOutput0[3], unsigned char *hOutput1[3], int width, int height){

	imgWidth = width;
	imgHeight = height;

	currentHostBuffer = 0;

	extern unsigned short *kduOutputHost[3];
	extern unsigned short *kduOutputDevice[3];

	hostOutput0[0] = hOutput0[0];
	hostOutput0[1] = hOutput0[1];
	hostOutput0[2] = hOutput0[2];

	hostOutput1[0] = hOutput1[0];
	hostOutput1[1] = hOutput1[1];
	hostOutput1[2] = hOutput1[2];

	deviceInput[0] = (unsigned char *) AllocateMemDevice(imgWidth*imgHeight*2);
	deviceInput[1] = (deviceInput[0]+imgWidth*imgHeight);
	deviceInput[2] = (deviceInput[0]+imgWidth*imgHeight*3/2);

	deviceIntermediate0[0] = (unsigned char *) AllocateMemDevice(imgWidth*imgHeight*2);
	deviceIntermediate0[1] = (deviceIntermediate0[0]+imgWidth*imgHeight);
	deviceIntermediate0[2] = (deviceIntermediate0[0]+imgWidth*imgHeight*3/2);

	deviceIntermediate1[0] = (short *) AllocateMemDevice(imgWidth*imgHeight*4);
	deviceIntermediate1[1] = (deviceIntermediate1[0]+imgWidth*imgHeight);
	deviceIntermediate1[2] = (deviceIntermediate1[0]+imgWidth*imgHeight*3/2);

	hostOutput0Mutex = OpenMutex(STANDARD_RIGHTS_ALL, false,(LPCTSTR)"hostOutput0Mutex");
	if(hostOutput0Mutex==NULL){
		DWORD error = GetLastError();
		fprintf(stderr, "Failed opening hostOutput0Mutex with error code: 0x%x\n", error);
		exit(-1);
	}

	hostOutput1Mutex = OpenMutex(STANDARD_RIGHTS_ALL, false,(LPCTSTR)"hostOutput1Mutex");
	if(hostOutput1Mutex==NULL){
		DWORD error = GetLastError();
		fprintf(stderr, "Failed opening hostOutput1Mutex with error code: 0x%x\n", error);
		exit(-1);
	}

    // initialize GLUT
	char *argv2[] = {"InitGui"};
	int argc2 = 1;
    glutInit(&argc2, argv2);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowSize(width, height);
    glutCreateWindow("JPEG2000 Video Decoder");
    glutDisplayFunc(Display);
    glutReshapeFunc(Reshape);
    //glutIdleFunc(Idle);
    glewInit();

	if (!glewIsSupported( "GL_VERSION_1_5 GL_ARB_vertex_buffer_object GL_ARB_pixel_buffer_object" )) {
		fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
		fprintf(stderr, "This sample requires:\n");
		fprintf(stderr, "  OpenGL version 1.5\n");
		fprintf(stderr, "  GL_ARB_vertex_buffer_object\n");
		fprintf(stderr, "  GL_ARB_pixel_buffer_object\n");
		exit(-1);
	}

    InitOpenGL();
    //atexit(CleanupGL);
}//end InitGui()

//Method called from main thread to start GLUT mainloop
//Will not return before GUI window is closed.
//Parameters:
//
//Return: void
extern "C" void StartGuiMainloop(){

	//Mapping PBO to CUDA memory space
	//cutilSafeCall(cudaGLMapBufferObject((void**)&mappedPbo, pbo));

	//Entering the glut main loop, which will not return before the window is closed.
    glutMainLoop();

	//Unmapping PBO from CUDA memory space
	//cutilSafeCall(cudaGLUnmapBufferObject(pbo));

	//Cleanup when then window is closed
	FreeMem(hostOutput0[0]);
	FreeMem(hostOutput1[0]);
	FreeMem(deviceInput[0]);
	CleanupGL();
    cudaThreadExit();
	fprintf(stderr, "glutMainLoop() done.\n");
}//end StartGuiMainloop()