/**********************************************************************************

	AUTHOR: Erlend Halsteinli
	CREATED: SPRING 2009 @ IET NTNU

	DWT coefficients are interleaved within the kanvas, as decribed in chapter 10.3.2 in the Taubman JPEG2000 book.
	Depending on the consuming method in the reverse wavelet transform, the coefficients must be deinterleaved.

***********************************************************************************/

#ifndef DEINTERLEAVE_H
#define DEINTERLEAVE_H

enum ROW_TYPE{
	LL1,	//Used if dwtLevel is 1
	LL5,	//Use if dwtLevel is 5
	LH1,
	LH2,
	LH3,
	LH4,
	LH5,
	ERR
};

//Prototypes
__device__ enum ROW_TYPE BacktrackRowtype(int rowIdx);
__global__ void DeinterleaveVisual(ushort4 *interleavedH, unsigned char *deinterleavedH, int imgWidth, int Height, int scale, int dwtLevel);
__global__ void DeinterleaveLinear(ushort4 *interleavedH, unsigned char *deinterleavedH, int imgWidth, int Height, int scale, int dwtLevel);


//Method performing DWT coefficients deinterleaving, either "linear" or "visual".
//dwtLevel must be one of: {1,5}
void Deinterleave(ushort4 *interleavedH, unsigned char *deinterleavedH, int imgWidth, int imgHeight, int scale, int dwtLevel){


	//Creating thread block and grid dimensions
	dim3 deinterleaveBlock(imgWidth/8,1,1);
	dim3 deinterleaveGrid(1,imgHeight,1);
	//Kernel launch.
	DeinterleaveVisual<<< deinterleaveGrid, deinterleaveBlock >>>((ushort4*)interleavedH, deinterleavedH, imgWidth, imgHeight, scale, dwtLevel);
	//DeinterleaveLinear<<< deinterleaveGrid, deinterleaveBlock >>>((ushort4*)interleavedH, deinterleavedH, imgWidth, imgHeight, scale, dwtLevel);
	//cutilCheckMsg("DeinterleaveVisual kernel failed");
}


//Deinterleaving into "visual" representation, i.e. LL band in upper left corner, HL in upper right, LH in bottom left and HH in bottom right.
//When process one component, each block is processing one row. Invoke therefore with block(imgWidth/8,1,1) and grid(1,imgHeight,1)
__global__ void DeinterleaveVisual(ushort4 *interleavedH, unsigned char *deinterleavedH, int imgWidth, int imgHeight, int scale, int dwtLevel){

	enum ROW_TYPE rowType;

	//First check if current row is odd or even.
	if( ( blockIdx.y % 2 ) == 0){
		//Even. Must determin the row type, by back tracking the row index. Row type is identified by the type if coefficients it contains. 
		//We know the interval betwen each row type, and where each row type first appeared. Use this to find the row type.
		if(dwtLevel==5)
			rowType = BacktrackRowtype(blockIdx.y);
		else if (dwtLevel==1)
			rowType = LL1;
		else
			rowType = ERR;
	}
	else{
		//Odd. All rows with odd index is of type LH1.
		rowType=LH1;
	}

	//Different rows contain differnet DWT coeffs. Based on rowType wee know what kind of coeffs the current row contains.
	//All treads within a block process the same row and thus have the same rowType. This is done in order to avoid conditional branching within a block,
	//since conditional branching within a block(more precise: within a warp) is a performance killer.
	ushort4 tmp;
	switch(rowType){
		//Use all threads in block
		case LL1:
			//Reading section of row contains LL1-HL1-LL1-HL1-LL1-HL1-LL1-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];
			//Writing LL1 samples
			deinterleavedH[(blockIdx.y)*imgWidth/2 + 4*threadIdx.x+0]=(tmp.x & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/2 + 4*threadIdx.x+1]=(tmp.y & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/2 + 4*threadIdx.x+2]=(tmp.z & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/2 + 4*threadIdx.x+3]=(tmp.w & 0xFF);
			//Writing HL1 samples
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH1:
			//Reading section of row contains LH1-HH1-LH1-HH1-LH1-HH1-LH1-HH1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH1 samples
			//imgWidth*imgHeight/2 - top left corner of LH1 subband.
			//(blockIdx.y-1)*imgWidth/2 - offset from top left corner for current row.
			//4*threadIdx.x+a - offset into current row
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+1]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
			//Writing HH1 samples
			//imgWidth*imgHeight/2+imgWidth/2 - offset from top left corner for current row.
			deinterleavedH[imgWidth*imgHeight/2 + imgWidth/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + imgWidth/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + imgWidth/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + imgWidth/2 + (blockIdx.y-1)*imgWidth/2 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH2:
			//Reading section of row contains LH2-HL1-HH2-HL1-LH2-HL1-HH2-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH2 samples
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y-2)*imgWidth/4 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y-2)*imgWidth/4 + 2*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
			//Writing HH2 samples
			deinterleavedH[imgWidth*imgHeight/4 + imgWidth/4 + (blockIdx.y-2)*imgWidth/4+2*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + imgWidth/4 + (blockIdx.y-2)*imgWidth/4+2*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
			//Writing HL1 samples
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH3:
			//Reading section of row contains LH3-HL1-HL2-HL1-HH3-HL1-HL2-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH3 sample
			deinterleavedH[imgWidth*imgHeight/8 + (blockIdx.y-4)*imgWidth/8 + threadIdx.x]=(tmp.x & 0xFF)-scale;
			//Writing HL1 samples
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			//Writing HL2 samples
			deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 2*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 2*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
			//Writing HH3 sample
			deinterleavedH[imgWidth*imgHeight/8 + imgWidth/8 + (blockIdx.y-4)*imgWidth/8 + threadIdx.x]=(tmp.z & 0xFF)-scale;

			break;
		//Use half of all threads in block
		case LH4:
			if(threadIdx.x < (blockDim.x/2)){
				//Reading section of row contains LH4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+2*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 2*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LH4 sample
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y-8)*imgWidth/16 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HH4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+2*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+4]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+5]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+6]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 8*threadIdx.x+7]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 2*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HH4 sample
				deinterleavedH[imgWidth*imgHeight/16 + imgWidth/16 + (blockIdx.y-8)*imgWidth/16 + threadIdx.x]=(tmp.z & 0xFF)-scale;
			}
			break;
		//Use one fourth of all threads in block
		case LH5:
			if(threadIdx.x < (blockDim.x/4)){
				//Reading section of row contains LH5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+0]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+1]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+2]= ((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+3]= ((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LH5 sample
				deinterleavedH[imgWidth*imgHeight/32 + imgWidth/2 + (blockIdx.y-16)*imgWidth/32 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+4]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+5]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+6]= ((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+7]= ((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+1]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HH5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+2];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+8]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+9]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+10]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+11]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+4]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+5]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
				//Writing HH5 sample
				deinterleavedH[imgWidth*imgHeight/32 + imgWidth/2 + imgWidth/32 + (blockIdx.y-16)*imgWidth/32 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+3];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+12]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+13]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+14]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+15]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+6]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+7]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+3]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			}
			break;
		//Use one fourth of all threads in block
		case LL5:
			if(threadIdx.x < (blockDim.x/4)){
				//Reading section of row contains LL5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+0]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+1]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+2]= ((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+3]= ((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LL5 sample
				deinterleavedH[(blockIdx.y)*imgWidth/32 + threadIdx.x]= (tmp.x & 0xFF);

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+4]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+5]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+6]= ((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+7]= ((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+2];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+8]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+9]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+10]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+11]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+4]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+5]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
				//Writing HL5 samples
				deinterleavedH[imgWidth/32 + (blockIdx.y)*imgWidth/32 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+3];
				//Writing HL1 samples
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+12]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+13]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+14]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth/2 + (blockIdx.y)*imgWidth/2 + 16*threadIdx.x+15]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+6]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+7]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth/8 + (blockIdx.y)*imgWidth/8 + 4*threadIdx.x+3]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+1]=(tmp.x & 0xFF)-scale;
			}
			break;
		case ERR:
			break;
	}

}//end DeinterleaveVisual()

//Deinterleaving into "linear" representation, i.e. LL5, HL5, LH5, HH5, HL4, LH4, HH4, HL3, LH3, HH3, HL2, LH2, HH2, HL1, LH1, HH1.
//When process one component, each block is processing one row. Invoke therefore with block(imgWidth/8,1,1) and grid(1,imgHeight,1)
__global__ void DeinterleaveLinear(ushort4 *interleavedH, unsigned char *deinterleavedH, int imgWidth, int imgHeight, int scale, int dwtLevel){

	enum ROW_TYPE rowType;

	//First check if current row is odd or even.
	if( ( blockIdx.y % 2 ) == 0){
		//Even. Must determin the row type, by back tracking the row index. Row type is identified by the type if coefficients it contains. 
		//We know the interval betwen each row type, and where each row type first appeared. Use this to find the row type.
		if(dwtLevel==5)
			rowType = BacktrackRowtype(blockIdx.y);
		else if (dwtLevel==1)
			rowType = LL1;
		else
			rowType = ERR;
	}
	else{
		//Odd. All rows with odd index is of type LH1.
		rowType=LH1;
	}

	//Different rows contain differnet DWT coeffs. Based on rowType wee know what kind of coeffs the current row contains.
	//All treads within a block process the same row and thus have the same rowType. This is done in order to avoid conditional branching within a block,
	//since conditional branching within a block(more precise: within a warp) is a performance killer.
	ushort4 tmp;
	switch(rowType){
		//Use all threads in block
		case LL1:
			//Reading section of row contains LL1-HL1-LL1-HL1-LL1-HL1-LL1-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];
			//Writing LL1 samples
			deinterleavedH[(blockIdx.y)*imgWidth/4 + 4*threadIdx.x+0]=(tmp.x & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/4 + 4*threadIdx.x+1]=(tmp.y & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/4 + 4*threadIdx.x+2]=(tmp.z & 0xFF);
			deinterleavedH[(blockIdx.y)*imgWidth/4 + 4*threadIdx.x+3]=(tmp.w & 0xFF);
			//Writing HL1 samples
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH1:
			//Reading section of row contains LH1-HH1-LH1-HH1-LH1-HH1-LH1-HH1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH1 samples
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+1]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/2 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
			//Writing HH1 samples
			deinterleavedH[imgWidth*imgHeight*3/4 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight*3/4 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight*3/4 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight*3/4 + (blockIdx.y-1)*imgWidth/4 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH2:
			//Reading section of row contains LH2-HL1-HH2-HL1-LH2-HL1-HH2-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH2 samples
			deinterleavedH[imgWidth*imgHeight/8 + (blockIdx.y-2)*imgWidth/16 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/8 + (blockIdx.y-2)*imgWidth/16 + 2*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
			//Writing HH2 samples
			deinterleavedH[imgWidth*imgHeight*3/16 + (blockIdx.y-2)*imgWidth/16 + 2*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight*3/16 + (blockIdx.y-2)*imgWidth/16 + 2*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
			//Writing HL1 samples
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			break;
		//Use all threads in block
		case LH3:
			//Reading section of row contains LH3-HL1-HL2-HL1-HH3-HL1-HL2-HL1.
			tmp = interleavedH[blockIdx.y*blockDim.x+threadIdx.x];

			//Writing LH3 sample
			deinterleavedH[imgWidth*imgHeight/32 + (blockIdx.y-4)*imgWidth/64 + threadIdx.x]=(tmp.x & 0xFF)-scale;
			//Writing HL1 samples
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 4*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
			//Writing HL2 samples
			deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
			deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 2*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
			//Writing HH3 sample
			deinterleavedH[imgWidth*imgHeight*3/64 + (blockIdx.y-4)*imgWidth/64 + threadIdx.x]=(tmp.z & 0xFF)-scale;

			break;
		//Use half of all threads in block
		case LH4:
			if(threadIdx.x < (blockDim.x/2)){
				//Reading section of row contains LH4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+2*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 4*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 4*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 2*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LH4 sample
				deinterleavedH[imgWidth*imgHeight/128 + (blockIdx.y-8)*imgWidth/256 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HH4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+2*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+4]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+5]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+6]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 8*threadIdx.x+7]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 4*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 4*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 2*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HH4 sample
				deinterleavedH[imgWidth*imgHeight*3/256 + (blockIdx.y-8)*imgWidth/256 + threadIdx.x]=(tmp.z & 0xFF)-scale;
			}
			break;
		//Use one fourth of all threads in block
		case LH5:
			if(threadIdx.x < (blockDim.x/4)){
				//Reading section of row contains LH5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LH5 sample
				deinterleavedH[(imgWidth*imgHeight)/512+imgWidth/32 + ((blockIdx.y-16)*imgWidth)/1024 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+4]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+5]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+6]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+7]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth*imgHeight/256 + (blockIdx.y)*imgWidth/256 + 2*threadIdx.x+1]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HH5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+2];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+8]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+9]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+10]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+11]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+4]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+5]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
				//Writing HH5 sample
				deinterleavedH[(imgWidth*imgHeight*3)/1024+imgWidth/64 + ((blockIdx.y-16)*imgWidth)/1024 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+3];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+12]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+13]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+14]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+15]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+6]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+7]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+3]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth*imgHeight/256 + (blockIdx.y)*imgWidth/256 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;
			}
			break;
		//Use one fourth of all threads in block
		case LL5:
			if(threadIdx.x < (blockDim.x/4)){
				//Reading section of row contains LL5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+0];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+0]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+1]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+2]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+3]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+0]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+1]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+0]=(tmp.z & 0xFF)-scale;
				//Writing LL5 sample
				deinterleavedH[((blockIdx.y)*imgWidth)/1024 + threadIdx.x]=(tmp.x & 0xFF);

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+1];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+4]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+5]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+6]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+7]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+2]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+3]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+1]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth*imgHeight/256 + (blockIdx.y)*imgWidth/256 + 2*threadIdx.x+0]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL5-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+2];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+8]= ((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+9]= ((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+10]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+11]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+4]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+5]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+2]=(tmp.z & 0xFF)-scale;
				//Writing HL5 samples
				deinterleavedH[(imgWidth*imgHeight)/1024+imgWidth/64 + ((blockIdx.y)*imgWidth)/1024 + threadIdx.x]=(tmp.x & 0xFF)-scale;

				//Reading section of row contains HL4-HL1-HL2-HL1-HL3-HL1-HL2-HL1
				tmp = interleavedH[blockIdx.y*blockDim.x+4*threadIdx.x+3];
				//Writing HL1 samples
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+12]=((tmp.x >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+13]=((tmp.y >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+14]=((tmp.z >> 8) & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/4 + (blockIdx.y)*imgWidth/4 + 16*threadIdx.x+15]=((tmp.w >> 8) & 0xFF)-scale;
				//Writing HL2 samples
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+6]=(tmp.y & 0xFF)-scale;
				deinterleavedH[imgWidth*imgHeight/16 + (blockIdx.y)*imgWidth/16 + 8*threadIdx.x+7]=(tmp.w & 0xFF)-scale;
				//Writing HL3 samples
				deinterleavedH[imgWidth*imgHeight/64 + (blockIdx.y)*imgWidth/64 + 4*threadIdx.x+3]=(tmp.z & 0xFF)-scale;
				//Writing HL4 samples
				deinterleavedH[imgWidth*imgHeight/256 + (blockIdx.y)*imgWidth/256 + 2*threadIdx.x+1]=(tmp.x & 0xFF)-scale;
			}
			break;
		case ERR:
			break;
	}

}//end DeinterleaveLinear()

//A row contains differnet DWT coeffs determined by the "row type". The row type is found from the row index(which is equal to the blockIdx.y), by backtracking.
__device__ enum ROW_TYPE BacktrackRowtype(int rowIdx){

	bool rowTypeFound=false;
	int stepSize = 32, k=0, targetIdx=0;
	enum ROW_TYPE rowCandidate = LL5;
	while(!rowTypeFound){
		while(blockIdx.y >= k*stepSize){
			//Found correct rowType, done searching
			if((blockIdx.y - k*stepSize)==targetIdx){
				rowTypeFound=true;
				return rowCandidate;
			}
			//try another step with same rowType candidate. Candidate given by 
			else{
				k++;
			}
		}
		switch(rowCandidate){
			//row was not LL5, try LH5
			case LL5:
				rowCandidate=LH5;
				k=0;
				stepSize=32;
				targetIdx=16;
				break;
			//row was not LH5, try LH4
			case LH5:
				rowCandidate=LH4;
				k=0;
				stepSize=16;
				targetIdx=8;
				break;
			//row was not LH4, try LH2
			case LH4:
				rowCandidate=LH3;
				k=0;
				stepSize=8;
				targetIdx=4;
				break;
			//row was not LH3, try LH2
			case LH3:
				rowCandidate=LH2;
				k=0;
				stepSize=4;
				targetIdx=2;
				break;
			//something went wrong
			case LH2:
				return ERR;
		}

	}

	return ERR;
}//end BacktrackRowtype()

#endif