/**********************************************************************************

	AUTHOR: Erlend Halsteinli
	CREATED: SPRING 2009 @ IET NTNU

	CUDA kernel performing YCbCr to RGBA color space transform.

***********************************************************************************/

#ifndef YCC2RGBA_KERNEL_H
#define YCC2RGBA_KERNEL_H

//Prototypes
__global__ void Ycc2Rgba16(uint4 *rgbaImage, ushort4 *yComponent, ushort2 *cbComponent, ushort2 *crComponent);
__global__ void Ycc2Rgba8(uint4 *rgbaImage, uchar4 *yComponent, uchar2 *cbComponent, uchar2 *crComponent);

void Ycc2Rgba(unsigned int* rgbaImage, unsigned char* yComponent, unsigned char* cbComponent, unsigned char* crComponent, int imgWidth, int imgHeight){

	//Creating thread block and grid dimensions
	dim3 ictBlock(8, 8, 1);
	dim3 ictGrid(imgWidth / ictBlock.x / 4, imgHeight / ictBlock.y, 1);
	//Kernel launch.
	Ycc2Rgba8<<< ictGrid, ictBlock >>>((uint4*)rgbaImage, (uchar4*)yComponent, (uchar2*)cbComponent, (uchar2*)crComponent);
	//Ycc2Rgba16<<< ictGrid, ictBlock >>>((uint4*)rgbaImage, (ushort4*)yComponent, (ushort2*)cbComponent, (ushort2*)crComponent);
	//cutilCheckMsg("Ycc2Rgba kernel failed");
}//end Ycc2Rgba()

//Input values are asumed to have 16 bit per sample per components(NB: only top byte per sample is use in calculations), while output have 32bpp.
__global__ void Ycc2Rgba16(uint4 *rgbaImage, ushort4 *yComponent, ushort2 *cbComponent, ushort2 *crComponent){

	unsigned idx = threadIdx.x + threadIdx.y*blockDim.x + blockDim.x*blockDim.y*(blockIdx.x +blockIdx.y*gridDim.x);

	ushort4 yTemp  =  yComponent[idx];
	ushort2 cbTemp = cbComponent[idx];
	ushort2 crTemp = crComponent[idx];

	//0xABGR because of byte order on graphics hardware
	uint4 result = make_uint4(0, 0, 0, 0);

	result.x = 0xFF000000 
			| ((unsigned char)((float)(yTemp.x & 0xFF) + 1.732f*((float)(cbTemp.x & 0xFF)-128)) << 16)
			| ((unsigned char)((float)(yTemp.x & 0xFF) - 0.336f*((float)(cbTemp.x & 0xFF)-128) - 0.698f*((float)(crTemp.x & 0xFF)-128)) << 8 )
			| (unsigned char)((float)(yTemp.x & 0xFF) + 1.371f*((float)(crTemp.x)-128));

	result.y = 0xFF000000 
			| ((unsigned char)((float)(yTemp.y & 0xFF) + 1.732f*((float)(cbTemp.x & 0xFF)-128)) << 16)
			| ((unsigned char)((float)(yTemp.y & 0xFF) - 0.336f*((float)(cbTemp.x & 0xFF)-128) - 0.698f*((float)(crTemp.x & 0xFF)-128)) << 8 )
			| (unsigned char)((float)(yTemp.y & 0xFF) + 1.371f*((float)(crTemp.x)-128));

	result.z = 0xFF000000 
			| ((unsigned char)((float)(yTemp.z & 0xFF) + 1.732f*((float)(cbTemp.y & 0xFF)-128)) << 16 )
			| ((unsigned char)((float)(yTemp.z & 0xFF) - 0.336f*((float)(cbTemp.y & 0xFF)-128) - 0.698f*((float)(crTemp.y & 0xFF)-128)) << 8)
			| (unsigned char)((float)(yTemp.z & 0xFF) + 1.371f*((float)(crTemp.y)-128));

	result.w = 0xFF000000 
			| ((unsigned char)((float)(yTemp.w & 0xFF) + 1.732f*((float)(cbTemp.y & 0xFF)-128)) << 16)
			| ((unsigned char)((float)(yTemp.w & 0xFF) - 0.336f*((float)(cbTemp.y & 0xFF)-128) - 0.698f*((float)(crTemp.y & 0xFF)-128)) << 8)
			| (unsigned char)((float)(yTemp.w & 0xFF) + 1.371f*((float)(crTemp.y)-128));

	rgbaImage[idx] =  result;
}//end Ycc2Rgba16()

__global__ void Ycc2Rgba8(uint4 *rgbaImage, uchar4 *yComponent, uchar2 *cbComponent, uchar2 *crComponent){

	unsigned idx = threadIdx.x + threadIdx.y*blockDim.x + blockDim.x*blockDim.y*(blockIdx.x +blockIdx.y*gridDim.x);

	uchar4 yTemp  =  yComponent[idx];
	uchar2 cbTemp = cbComponent[idx];
	uchar2 crTemp = crComponent[idx];
	float y, cb, cr;
	//y should be in intervall: [16,235], but noise and round-off errors may give values outside.
	//cb, cr should be in intervall: [16,240], but noise and round-off errors may give values outside.

	//0xABGR because of byte order on graphics hardware
	uint4 result = make_uint4(0, 0, 0, 0);

	y = (yTemp.x > 235) ? 219 : yTemp.x-16;
	y = (y < 0) ? 0 : y/219;
	//y=pow(y,1.2f);
	cb = (cbTemp.x > 240) ? 112 : cbTemp.x-128;
	cb = (cb < -112) ? -0.5 : cb/224;
	cr = (crTemp.x > 240) ? 112 : crTemp.x-128;
	cr = (cr < -112) ? -0.5 : cr/224;

	result.x = 0xFF000000
			| (((unsigned char)((y + 1.772f*cb)*255)) << 16)
			| (((unsigned char)((y - 0.344f*cb - 0.714f*cr)*255)) << 8 )
			|  (unsigned char)((y + 1.402f*cr)*255);

	y = (yTemp.y > 235) ? 219 : yTemp.y-16;
	y = (y < 0) ? 0 : y/219;
	//y=pow(y,1.2f);

	result.y = 0xFF000000
			| (((unsigned char)((y + 1.772f*cb)*255)) << 16)
			| (((unsigned char)((y - 0.344f*cb - 0.714f*cr)*255)) << 8 )
			|  (unsigned char)((y + 1.402f*cr)*255);

	y = (yTemp.z > 235) ? 219 : yTemp.z-16;
	y = (y < 0) ? 0 : y/219;
	//y=pow(y,1.2f);

	cb = (cbTemp.y > 240) ? 112 : cbTemp.y-128;
	cb = (cb < -112) ? -0.5 : cb/224;
	cr = (crTemp.y > 240) ? 112 : crTemp.y-128;
	cr = (cr < -112) ? -0.5 : cr/224;

	result.z = 0xFF000000
			| (((unsigned char)((y + 1.772f*cb)*255)) << 16)
			| (((unsigned char)((y - 0.344f*cb - 0.714f*cr)*255)) << 8 )
			|  (unsigned char)((y + 1.402f*cr)*255);

	y = (yTemp.w > 235) ? 219 : yTemp.w-16;
	y = (y < 0) ? 0 : y/219;
	//y=pow(y,1.2f);

	result.w = 0xFF000000
			| (((unsigned char)((y + 1.772f*cb)*255)) << 16)
			| (((unsigned char)((y - 0.344f*cb - 0.714f*cr)*255)) << 8 )
			|  (unsigned char)((y + 1.402f*cr)*255);

	//result.x = 0xFF000000 | (yTemp.x<<16) | (yTemp.x<<8) | yTemp.x;
	//result.y = 0xFF000000 | (yTemp.y<<16) | (yTemp.y<<8) | yTemp.y;
	//result.z = 0xFF000000 | (yTemp.z<<16) | (yTemp.z<<8) | yTemp.z;
	//result.w = 0xFF000000 | (yTemp.w<<16) | (yTemp.w<<8) | yTemp.w;

	//result.x =  (yTemp.x<<24) | (yTemp.x<<16) | (yTemp.x<<8) | 0xFF;
	//result.y =  (yTemp.y<<24) | (yTemp.y<<16) | (yTemp.y<<8) | 0xFF;
	//result.z =  (yTemp.z<<24) | (yTemp.z<<16) | (yTemp.z<<8) | 0xFF;
	//result.w =  (yTemp.w<<24) | (yTemp.w<<16) | (yTemp.w<<8) | 0xFF;

	//result.x = 0xFF000000 | (cbTemp.x<<16) | (cbTemp.x<<8) | cbTemp.x;
	//result.y = 0xFF000000 | (cbTemp.x<<16) | (cbTemp.x<<8) | cbTemp.x;
	//result.z = 0xFF000000 | (cbTemp.y<<16) | (cbTemp.y<<8) | cbTemp.y;
	//result.w = 0xFF000000 | (cbTemp.y<<16) | (cbTemp.y<<8) | cbTemp.y;

	//result.x = 0xFF000000 | (crTemp.x<<16) | (crTemp.x<<8) | crTemp.x;
	//result.y = 0xFF000000 | (crTemp.x<<16) | (crTemp.x<<8) | crTemp.x;
	//result.z = 0xFF000000 | (crTemp.y<<16) | (crTemp.y<<8) | crTemp.y;
	//result.w = 0xFF000000 | (crTemp.y<<16) | (crTemp.y<<8) | crTemp.y;

	rgbaImage[idx] =  result;
}//end Ycc2Rgba8()

//convert from unsigned char to signed short
__global__ void char2short(uchar4 *input, short4* output){

	unsigned idx = threadIdx.x + threadIdx.y*blockDim.x + blockDim.x*blockDim.y*(blockIdx.x +blockIdx.y*gridDim.x);

	uchar4 in = input[idx];
	short4 out = make_short4(0,0,0,0);
	out.x = (((char)in.x - 0x80) << 8);
	out.y = (((char)in.y - 0x80) << 8);
	out.z = (((char)in.z - 0x80) << 8);
	out.w = (((char)in.w - 0x80) << 8);
	output[idx] = out;
}//end char2short()

//convert from signed short to unsigned char
__global__ void short2char(short4* input, uchar4 *output){

	unsigned idx = threadIdx.x + threadIdx.y*blockDim.x + blockDim.x*blockDim.y*(blockIdx.x +blockIdx.y*gridDim.x);

	short4 in = input[idx];
	uchar4 out = make_uchar4(0,0,0,0);
	out.x = ((in.x >> 8) & 0xFF) + 0x80;
	out.y = ((in.y >> 8) & 0xFF) + 0x80;
	out.z = ((in.z >> 8) & 0xFF) + 0x80;
	out.w = ((in.w >> 8) & 0xFF) + 0x80;
	output[idx] = out;
	//yComponent[idx]= ((yComponent[idx] << 8) | 0xFF);

}//end short2char()
#endif
